/*****************************************************************************
 * vim:tw=78
 *
 * Copyright (C) 2021 Bjarne von Horn (vh at igh dot de).
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

/** @file */

#ifndef PDCOM5_SECUREPROCESS_H
#define PDCOM5_SECUREPROCESS_H

#include <memory>
#include <pdcom5/Process.h>

#if defined _WIN32 || defined __CYGWIN__
#ifdef pdcom5_gnutls_EXPORTS
#define PDCOM5_GNUTLS_PUBLIC __declspec(dllexport)
#else
#define PDCOM5_GNUTLS_PUBLIC __declspec(dllimport)
#endif
#else
#define PDCOM5_GNUTLS_PUBLIC __attribute__((visibility("default")))
#endif

namespace PdCom {

/** Process implementation for TLS encrypted traffic.
 *
 * \example gnutls_example.cpp
 *
 */
class PDCOM5_GNUTLS_PUBLIC SecureProcess : public Process
{
  public:
    struct PDCOM5_GNUTLS_PUBLIC EncryptionDetails
    {
        enum Flags {
            Default = 0,
        } flags_;
        std::string server_ca_, server_hostname_, client_cert_, client_key_;

        EncryptionDetails(
                Flags flags,
                std::string server_ca,
                std::string hostname,
                std::string client_cert = "",
                std::string client_key  = "") :
            flags_(flags),
            server_ca_(server_ca),
            server_hostname_(hostname),
            client_cert_(client_cert),
            client_key_(client_key)
        {}
        EncryptionDetails(
                std::string server_ca,
                std::string hostname,
                std::string client_cert = "",
                std::string client_key  = "") :
            flags_(Default),
            server_ca_(server_ca),
            server_hostname_(hostname),
            client_cert_(client_cert),
            client_key_(client_key)
        {}
    };

    /** GnuTls global initialization.
     *
     * Call this at startup of your application to prepare the underlying TLS
     * library.
     *
     * \throws PdCom::Exception Initialization failed.
     */
    static void InitLibrary();
    /** GnuTls global finalization.
     */
    static void FinalizeLibrary();

    SecureProcess() = default;
    explicit SecureProcess(EncryptionDetails const &);
    SecureProcess(SecureProcess &&) noexcept;
    SecureProcess &operator=(SecureProcess &&) noexcept;

    void handshake();
    void bye();

  private:
    struct Impl;
    std::shared_ptr<Impl> secure_impl_;
    void write(const char *buf, size_t count) final;
    int read(char *buf, int count) final;
    void flush() final;
    virtual int secureRead(char *buf, int count)            = 0;
    virtual void secureWrite(const char *buf, size_t count) = 0;
};

}  // namespace PdCom

#endif  // PDCOM5_SECUREPROCESS_H
