/*****************************************************************************
 * vim:tw=78
 *
 * Copyright (C) 2021 Richard Hacker (lerichi at gmx dot net),
 *                    Florian Pose (fp at igh dot de),
 *                    Bjarne von Horn (vh at igh dot de).
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

/** @file */

#ifndef PDCOM5_SIMPLELOGINMANAGER_H
#define PDCOM5_SIMPLELOGINMANAGER_H


#if defined _WIN32 || defined __CYGWIN__
#ifdef pdcom5_sasl_EXPORTS
#define PDCOM5_SASL_PUBLIC __declspec(dllexport)
#else
#define PDCOM5_SASL_PUBLIC __declspec(dllimport)
#endif
#else
#define PDCOM5_SASL_PUBLIC __attribute__((visibility("default")))
#endif

#include <memory>
#include <pdcom5/Sasl.h>
#include <stdexcept>

extern "C" struct sasl_callback;

namespace PdCom {
namespace impl {
class SimpleLoginManager;
}  // namespace impl

class PDCOM5_SASL_PUBLIC SimpleLoginManager : public Sasl
{
    friend class impl::SimpleLoginManager;
    std::unique_ptr<impl::SimpleLoginManager> impl_;
    void loginReply(const char *mechlist, const char *serverData, int finished)
            override;

  public:
    /** Sasl global initialization.
     *
     * Call this at startup of your application to initialize the underlying
     * sasl library.
     *
     * \throws PdCom::Exception Initialization failed.
     */
    static void InitLibrary();
    /** Sasl global finalization
     */
    static void FinalizeLibrary();

    SimpleLoginManager(
            const char *remote_host,
            sasl_callback *additional_callbacks = nullptr);
    SimpleLoginManager(SimpleLoginManager &&) noexcept;
    virtual ~SimpleLoginManager();
    SimpleLoginManager &operator=(SimpleLoginManager &&) noexcept;


    /** Exception for callback cancelation.
     *
     * Throw this in one of the callbacks (like getAuthname()) to cancel the
     * current login step.
     */
    struct Cancel : std::exception
    {};

    /** Perform SASL login step.
     *
     * @param mech SASL mechanism
     * @param clientData Base64 encoded SASL output data to server
     *
     * Setting both \p mech and \p clientData to NULL will initate the
     * login process.
     *
     * Every call to login() is answered by a loginReply(), unless
     * login is not supported. When login is mandatory, loginReply()
     * will be called automatically.
     *
     * @return false if login is not supported
     */
    bool login();

    /** Logout from server
     */
    using Sasl::logout;
    /** Callback to get login name.
     */
    virtual std::string getAuthname() { throw Cancel(); }
    /** Callback to get password.
     */
    virtual std::string getPassword() { throw Cancel(); }
    virtual std::string
    getOption(const char * /*plugin_name*/, const char * /*option*/)
    {
        throw Cancel();
    }
    virtual std::string interact(
            unsigned long /*id*/,
            const char * /*challenge*/,
            const char * /*prompt*/,
            const char * /*defresult*/)
    {
        throw Cancel();
    }
    virtual void completed(bool success) = 0;
};

}  // namespace PdCom

#endif  // PDCOM5_SIMPLELOGINMANAGER_H
