/*****************************************************************************
 * vim:tw=78
 *
 * Copyright (C) 2021 Richard Hacker (lerichi at gmx dot net),
 *                    Florian Pose (fp at igh dot de),
 *                    Bjarne von Horn (vh at igh dot de).
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

/** @file */

#ifndef PDCOM5_VARIABLE_H
#define PDCOM5_VARIABLE_H

#include "SizeTypeInfo.h"
#include "details.h"
#include "visibility.h"

#include <istream>
#include <memory>
#include <ostream>
#include <stdint.h>
#include <string>
#include <vector>

namespace PdCom {
namespace impl {
class Variable;
}

/** PdCom Variable interface.
 *
 * This class represents a variable (signal or parameter). It contains
 * information about the datatype and the shape of a variable.
 *
 * If you default-construct a Variable instance, it does not contain any
 * information. Calling any other member function than empty() is forbidden. To
 * get a valid variable instance, call Process::find().
 *
 */
class PDCOM5_PUBLIC Variable
{
    friend class impl::Variable;

    explicit Variable(std::weak_ptr<const impl::Variable> pimpl) :
        pimpl_(std::move(pimpl))
    {}

  public:
    /** Constructs an empty variable.
     */
    Variable() = default;

    /** Write to a variable.
     *
     * This function is also asynchronous, so the server might still use the old
     * value when this function returns. This overload is for any container of
     * arithmetic types (e.g. std::vector<int>).
     * The variable must not be empty!
     *
     * \c T must be a container (must have a size() method and a value_type
     * typedef).
     *
     * \param data Data.
     * \param index Optional offset.  No bounds checking!
     *
     * \return false if the variable is read-only.
     */
    template <typename T>
    typename std::enable_if<!std::is_arithmetic<T>::value, bool>::type
    setValue(T const &data, size_t index = 0) const
    {
        return setValue(
                &data[0],
                details::TypeInfoTraits<typename T::value_type>::type_info.type,
                index, data.size() - index);
    }
    /** Write to a variable.
     *
     * This function is also asynchronous, so the server might still use the old
     * value when this function returns.
     * This overload is for any arithmetic type (int, double, etc.).
     * The variable must not be empty!
     *
     * \param data Data.
     * \param index Optional offset.  No bounds checking!
     *
     * \return false if the variable is read-only.
     */
    template <typename T>
    typename std::enable_if<std::is_arithmetic<T>::value, bool>::type
    setValue(T const &data, size_t index = 0) const
    {
        return setValue(
                &data, details::TypeInfoTraits<T>::type_info.type, index, 1);
    }

    /** Write to a variable.
     *
     * This function is also asynchronous, so the server might still use the old
     * value when this function returns.
     * This overload is for any arithmetic type (int, double, etc.).
     * The variable must not be empty!
     *
     * \param data Data.
     * \param index Optional offset.  No bounds checking!
     *
     * \return false if the variable is read-only.
     */
    template <typename T, size_t M, size_t N>
    bool setValue(const T (&data)[M][N]) const
    {
        return setValue(
                data, details::TypeInfoTraits<T>::type_info.type, 0, M * N);
    }

    /** Write to a variable.
     *
     * This function is also asynchronous, so the server might still use the old
     * value when this function returns.
     * The variable must not be empty!
     *
     * \param src Pointer to one or more values.
     * \param src_type Type of the source.
     * \param idx Optional offset.  No bounds checking!
     * \param n Number of values.
     *
     * \return false if the variable is read-only.
     */
    bool
    setValue(const void *src, TypeInfo::DataType src_type, size_t idx, size_t n)
            const;

    /** Get details about the variable type.
     * The variable must not be empty!
     * @return Variable type info.
     */
    TypeInfo getTypeInfo() const;
    /** Get details about the variable shape.
     * The variable must not be empty!
     * @return Variable shape info.
     */
    SizeInfo getSizeInfo() const;
    /** The Path of the variable.
     * The variable must not be empty!
     * @return The full path.
     */
    std::string getPath() const;
    /** The name of the variable.
     * The variable must not be empty!
     * @return the name, without the path components.
     */
    std::string getName() const;
    /** The alias of the variable, if set.
     * The variable must not be empty!
     * @return the alias.
     */
    std::string getAlias() const;
    /** The task id of the variable.
     * The variable must not be empty!
     * @return the task id.
     */
    int getTaskId() const;

    /** @return whether the variable is writeable.
     */
    bool isWriteable() const;


    /** Checks whether this instance is empty.
     *
     * Default-constructed variables are empty per default,
     * calling any other member function than empty() will crash.
     * \return false if this instance is alive.
     */
    bool empty() const noexcept { return (pimpl_.expired()); }

  private:
    std::weak_ptr<const impl::Variable> pimpl_;
};

}  // namespace PdCom

#endif  // PDCOM5_VARIABLE_H
